/*
 * codec_func.c
 *
 * implements the "analog" operating mode of DVfirmware. Direct i/o with
 * the speech codec AD73311 (w/o using AMBE-2020).
 *
 *
 *  Created on: 19.09.2009
 *      Author: DO1FJN, Jan Alte
 *
 *
 * This file is part of the DV-modem firmware (DVfirmware).
 * For general infomation about DVfirmware see "dv-main.c".
 *
 * DVfirmware is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * The DVfirmware is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 * Report:
 * 2010-03-20	Outputgain +3dB insteed +6dB -> better match to D-Star-Volume
 */

#include "codec_func.h"
#include "sinegen.h"

#include <math.h>
#include "dsp.h"
#include "intc.h"
#include "../hw_init.h"


#define	CODEC_PRIO	AVR32_INTC_INT0

#define CODEC_RXINTEN()	(AVR32_SSC.ier = AVR32_SSC_IER_TXRDY_MASK)
#define CODEC_TXINTEN()	(AVR32_SSC.ier = AVR32_SSC_IER_TXRDY_MASK)
#define	CODEC_INTDIS()	(AVR32_SSC.idr = 0xFFFF)

#define codec_inactive()	(AVR32_SSC.imr==0)
#define codec_active()		(AVR32_SSC.imr!=0)

#define SAMPLESPERSEC	32000L


typedef struct {
  dsp16_t vol, gain, tone_vol, tone_mod, modlimit;
  U8 sqlevel, mutestate;
  __int_handler rx_func, tx_func, tone_func;
} tcodecparam;


const U16 ad73311init[] = {
  0x8113, 0x82F9, 0x8310, 0x8001
};


// 0x8113 = Wr #1->13	B, SCLK-Div=0, MCLK-Div=1
// 0x82F9 = Wr #2->F9	C, Enable all (PowerUp)
// 0x8310 = Wr #3->00	D, InpGain=0dB, OutGain=+3dB, not Mute, no ADCModReset
// 0x8001 = Wr #0 > 1	A, Enable Data Mode

// globale Variablen:

tcodecstate CodecState;
tcodecparam CodecParam;

// Forwards:
void codec_stop(void);

dsp32_t rx_sample;

#define DeEmphTimeConst		750E-6
#define	DeEmphBoost		0.67
#define DeEmph_A1		exp(-1.0/SAMPLESPERSEC/DeEmphTimeConst)
#define DeEmph_B0		(DeEmphBoost/SAMPLESPERSEC/DeEmphTimeConst)
#define Deemphasis(var_x, new_x)	{ var_x = dsp32_op_mul(var_x, DSP32_Q(DeEmph_A1)) + dsp32_op_mul(new_x, DSP32_Q(DeEmph_B0)); }


#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ssc_std_receive(void) {
  S32	raw_adc;
  dsp32_t rx_new;
  AVR32_SSC.sr;					// ACK SSC-Status
  raw_adc = HFIN;				// Store last ADC
  adc_startconversion();			// Enable next AD Conversion
  AVR32_SSC.thr = __builtin_sats(rx_sample, 11, 16);	// Put older (filtered) 16bit-Sample to AD73311 Codec
  rx_new = CodecParam.vol *  ADC_SIGNEDFS(raw_adc);	// Convert 10bit unsigned in 16bit signed
  Deemphasis(rx_sample, rx_new);
}

#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ssc_raw_receive(void) {
  S32	raw_adc;
  AVR32_SSC.sr;					// ACK SSC-Status
  raw_adc = HFIN;				// Store last ADC
  adc_startconversion();			// Enable next AD Conversion
  // Ende raw ADC
  // Put 16bit-Sample to AD73311 Codec
  AVR32_SSC.thr = __builtin_sats((CodecParam.vol * ADC_SIGNEDFS(raw_adc)), 14, 16);
}



int txphase;

#define TXLPCoeffsSize		17

A_ALIGNED const dsp32_t TXLPCoeffs[TXLPCoeffsSize] = {
    DSP32_Q(  0.018768310546875),
    DSP32_Q(  0.001190185546875),
    DSP32_Q( -0.02972412109375),
    DSP32_Q( -0.050079345703125),
    DSP32_Q( -0.031005859375),
    DSP32_Q(  0.040130615234375),
    DSP32_Q(  0.1448974609375),
    DSP32_Q(  0.23907470703125),
    DSP32_Q(  0.27691650390625),
    DSP32_Q(  0.23907470703125),
    DSP32_Q(  0.1448974609375),
    DSP32_Q(  0.040130615234375),
    DSP32_Q( -0.031005859375),
    DSP32_Q( -0.050079345703125),
    DSP32_Q( -0.02972412109375),
    DSP32_Q(  0.001190185546875),
    DSP32_Q(  0.018768310546875)
};

#define TXPreCoeffsSize		9

A_ALIGNED const dsp32_t TXPreCoeffs[TXPreCoeffsSize] = {
  DSP32_Q( -0.058197021484375),
  DSP32_Q( -0.062530517578125),
  DSP32_Q( -0.065765380859375),
  DSP32_Q( -0.0677490234375),
  DSP32_Q(  0.8179931640625),
  DSP32_Q( -0.0677490234375),
  DSP32_Q( -0.065765380859375),
  DSP32_Q( -0.062530517578125),
  DSP32_Q( -0.058197021484375)
};


// *** Transmitter Buffer and Variables ****
#define MOD_OUT_SIZE		4
#define PRE_IN_SIZE		(TXPreCoeffsSize+MOD_OUT_SIZE-1)
#define MOD_IN_SIZE		(TXLPCoeffsSize+MOD_OUT_SIZE-1)

A_ALIGNED dsp32_t preemph_in[PRE_IN_SIZE];	// Eingangsbuffer für Preemphasis
A_ALIGNED dsp32_t codecmod_in[MOD_IN_SIZE];	// Eingangsbuffer für Modulator-LowPass
A_ALIGNED dsp32_t codecmod_out[MOD_OUT_SIZE];	// geGaußte DAC Daten


#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ssc_raw_transmit(void) {
  static U16 dac_val;			// Ausgabewert für Modulator
  dsp16_t sample;
  AVR32_SSC.sr;
  dac_transmit(dac_val);		// Alten Wert ausgeben
  sample = AVR32_SSC.rhr;
  AVR32_SSC.thr = 0x0000;		// Codec-Out ist null
  dac_val = __builtin_sats(codecmod_out[txphase], 2, 14)+0x2000;
  // MicGain Verstärkung / Abschwächung
  sample =  __builtin_sats((CodecParam.gain * sample), 12, 16);
  // Hard Limiter
  sample = min(max(sample, -CodecParam.modlimit), CodecParam.modlimit);
  // FIR-Low-Pass (Soften Limiter)
  codecmod_in[MOD_IN_SIZE-MOD_OUT_SIZE+txphase] = sample;
  txphase++;	// Next
  if (txphase==MOD_OUT_SIZE) {	// In-Buffer filled
    txphase = 0;
    dsp32_filt_fir(codecmod_out, codecmod_in, MOD_IN_SIZE, (dsp32_t *)TXLPCoeffs, TXLPCoeffsSize);
    //Test: dsp32_vect_copy(codecmod_out, codecmod_in+TXLPCoeffsSize-1, MOD_OUT_SIZE);
    dsp32_vect_copy(codecmod_in, codecmod_in+MOD_OUT_SIZE, MOD_IN_SIZE-MOD_OUT_SIZE);
  } // fi do FILTER
}


#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ssc_std_transmit(void) {
  static U16 dac_val;			// Ausgabewert für Modulator
  dsp16_t sample;
  int cnt;
  AVR32_SSC.sr;
  dac_transmit(dac_val);		// Alten Wert ausgeben
  sample = AVR32_SSC.rhr;
  AVR32_SSC.thr = 0x0000;		// Codec-Out ist null
  //sample =  __builtin_sats((CodecParam.gain * sample), 14, 16);
  preemph_in[PRE_IN_SIZE-MOD_OUT_SIZE+txphase] = sample * CodecParam.gain; // Speichern in PreBuffer
  dac_val = __builtin_sats(codecmod_out[txphase], 2, 14)+0x2000;
  txphase++;	// Next
  if (txphase==MOD_OUT_SIZE) {	// In-Buffer filled
    txphase = 0;
    dsp32_filt_fir(codecmod_in+TXLPCoeffsSize-1, preemph_in, PRE_IN_SIZE, (dsp32_t *)TXPreCoeffs, TXPreCoeffsSize);
    dsp32_vect_copy(preemph_in, preemph_in+MOD_OUT_SIZE, PRE_IN_SIZE-MOD_OUT_SIZE);
    for (cnt = TXLPCoeffsSize-1; cnt<MOD_IN_SIZE; cnt++) {
      // MicGain Verstärkung / Abschwächung:
      sample =  __builtin_sats(codecmod_in[cnt], 11, 16);
      // Hard Limiter:
      codecmod_in[cnt] =  min(max(sample, -CodecParam.modlimit), CodecParam.modlimit);
    }  // rof limit
    dsp32_filt_fir(codecmod_out, codecmod_in, MOD_IN_SIZE, (dsp32_t *)TXLPCoeffs, TXLPCoeffsSize);
    // Test: dsp16_vect_copy(codecmod_out, codecmod_in+TXLPCoeffsSize-1, MOD_OUT_SIZE);
    dsp32_vect_copy(codecmod_in, codecmod_in+MOD_OUT_SIZE, MOD_IN_SIZE-MOD_OUT_SIZE);
  } // fi do FILTER
}


#define MaxTonCnt	4
U32	tonesamplelen;			// Down-Zähler, wie lange Tongenerator arbeitet
tsinedescr Ton[MaxTonCnt];
int	NoOfTone;

#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ssc_std_multitone(void) {
    int cnt;
  static U16 dac_val;			// Ausgabewert für Modulator
  static dsp32_t tone_val;
  AVR32_SSC.sr;
  dac_transmit(dac_val);		// DualTone -> Modulator
  if (tonesamplelen > 0) {
    tonesamplelen--;
    tone_val = 0;
    for (cnt=0; cnt<NoOfTone; cnt++)
      tone_val += getsine(&Ton[cnt]);
    tone_val >>= NoOfTone;
  } else {
    tone_val = dsp32_op_mul(tone_val, DSP32_Q(0.8));
  }
  AVR32_SSC.thr = __builtin_sats(tone_val * CodecParam.tone_vol, 14, 16);
  // S16 Tone zum Mithören
  dac_val = __builtin_sats(tone_val * CodecParam.tone_mod, 16, 14)+0x2000;
}


volatile int ssc_length;
volatile U16 *ssc_data;

#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ssc_init_int(void) {
  if (AVR32_SSC.sr&AVR32_SSC_SR_TXRDY_MASK) {
    if (ssc_length>0) {
      AVR32_SSC.thr = *ssc_data++;
      ssc_length--;
    } else {
      CODEC_INTDIS();		// IRQs off
      if (CodecState > CODECoff) {
	if (NoOfTone > 0) {	// Call~/DTMF-Tonausgabe geschaltet
          INTC_register_interrupt(CodecParam.tone_func, AVR32_SSC_IRQ, CODEC_PRIO);
          CODEC_TXINTEN();	// TXReady
	} else switch(CodecState) {
          case CODECreceive:
	    if (CodecParam.mutestate == 0) {
	      CODEC_RXINTEN();
              INTC_register_interrupt(CodecParam.rx_func, AVR32_SSC_IRQ, CODEC_PRIO);
	    } else codec_stop();
	    break;
          case CODECtransmit:
    	    INTC_register_interrupt(CodecParam.tx_func, AVR32_SSC_IRQ, CODEC_PRIO);
	    CODEC_TXINTEN();	// TXReady
	    break;
          default:
            codec_stop();
            break;
        }
      } else codec_stop();
    } // esle finish codec-init
  } // fi tx rdy
}



void codec_stop(void) {
  CODEC_INTDIS();
  ssc_length = 0;
  AVR32_SSC.cr = 0x0202;		// Disable SSC TX (Bit 8) & RX
  cpld_config(CPLDdisabled);
  dac_exit();
  adc_exit();
  idle_timer_start();
}


void codec_start(void) {
  CODEC_INTDIS();		//off
  idle_timer_stop();
  cpld_config(CPLDcfg_codec);
  ssc_data   = (U16 *)ad73311init;
  ssc_length = sizeof(ad73311init);
  INTC_register_interrupt(&ssc_init_int, AVR32_SSC_IRQ, CODEC_PRIO);
  rx_sample = DSP32_Q(0.5);
  txphase = 0;
//  rxphase = 0;
  adc_init();
  dac_init();
  cpld_start();
  AVR32_SSC.sr;
  CODEC_TXINTEN();			// TXReady
  AVR32_SSC.cr  = 0x0101;		// Enable SSC TX (Bit 8) & RX
}


void codec_init(void) {
  if (CodecState > CODECoff) {
    codec_stop();
  } // fi
  CodecState = CODECnoinit;
  CodecParam.mutestate = 1;		// Mute at Init.
  AVR32_SSC.cr   = 0x00008000;		// Do a software reset
  cpld_config(CPLDdisabled);
  cpld_start();
  //testsine.phase = 0x08000000;	// 1kHz
  //dsp32_vect_zeropad(receiver_in, receiver_in_len, receiver_in_len);
  //dsp32_vect_zeropad(receiver_out, receiver_out_len, receiver_out_len);
  AVR32_SSC.cmr  = 1;
  AVR32_SSC.rcmr = SSC_RCMR_CODEC;
  AVR32_SSC.rfmr = SSC_RFMR_CODEC;
  AVR32_SSC.tcmr = SSC_TCMR_INI;
  AVR32_SSC.tfmr = SSC_TFMR_INI;	// 16 Bits
  CodecState = CODECoff;
  idle_timer_start();
}


void codec_exit(void) {
  CODEC_INTDIS();
  AVR32_SSC.cr = 0x0202;		// Disable SSC TX (Bit 8) & RX
  cpld_config(CPLDdisabled);
  dac_exit();
  adc_exit();
  CodecState = CODECoff;
}



void codec_register(__int_handler func) {
  CODEC_INTDIS();
  AVR32_SSC.sr;
  INTC_register_interrupt(func, AVR32_SSC_IRQ, CODEC_PRIO);
}



void codec_transmit(void) {
  if (CodecState > CODECnoinit) {
    if (codec_inactive()) {		// Ist Codec abgeschaltet?
      CodecState = CODECtransmit;
      codec_start();			// dadurch bei receive() erstmal stop()
    } else if (CodecState != CODECtransmit) {
      CodecState = CODECtransmit;
      if (NoOfTone==0) {		// keine Tonausgabe - also hörbares RX
        codec_register(CodecParam.tx_func);
	CODEC_TXINTEN();		// Run INT-Function
      } // fi
      // Bei Tonausgabe "nur" State auf Transmit schalten
    } // fi not xmit
  } // fi initialized
}


void codec_receive(void) {
  if (CodecState > CODECnoinit) {
    tcodecstate oldState = CodecState;
    CodecState = CODECreceive;	// State ändern
    if (codec_inactive()) {
      if (CodecParam.mutestate == 0)
	codec_start();
    } else {
      if (CodecParam.mutestate > 0) {	// Ist Codec stumm?
        codec_stop();
      } else if ((oldState != CODECreceive) && (NoOfTone==0)) {
        // Bei Tonausgabe "nur" State auf Receive schalten
        codec_register(CodecParam.rx_func);
        CODEC_RXINTEN();		// Run INT-Function
      } // else not mute
    } // esle inacti
  } // fi initialized
}


int codec_rxismute(void) {
  if (CodecState > CODECoff) {
    return (NoOfTone==0)&&((CodecState==CODECtransmit)||(CodecParam.mutestate>0));
  } else return (0);
}


void codec_mute(void) {
  if ((CodecState==CODECreceive) && (codec_active())) {
    codec_stop();		// Audio aus
  } // fi
  CodecParam.mutestate++;
}


void codec_unmute(void) {
  CodecParam.mutestate = 0;
  if ((CodecState > CODECoff)&&(codec_inactive())) {
    codec_start();
  }
}


void codec_gentone(int freq1, int freq2, int ms) {
  if (CodecState > CODECnoinit) {
    tonesamplelen = 0;	// old tone off
    Ton[0].accu = 0;
    Ton[0].phase = ((U64)freq1<<32) / SAMPLESPERSEC;	//234881024;
    NoOfTone = 1;
    if (freq2 > 50) {
      Ton[1].accu = 0;	// 0° Start
      Ton[1].phase = ((U64)freq2<<32) / SAMPLESPERSEC;
      NoOfTone = 2;
    }
    if (ms > 0) {
      tonesamplelen = (SAMPLESPERSEC * ms + 500) / 1000;
    } else {
      tonesamplelen = 0xffffffff;
    }
    if (codec_inactive()) {	// Codec abgeschaltet?
      codec_start();
    } else {	// Codec ist aktiv
      codec_register(CodecParam.tone_func);
      CODEC_TXINTEN();		// Run INT-Function
    } // else
  } // fi init
}


void codec_calltone(void) {
  if (CodecState > CODECnoinit) {
    tonesamplelen = SAMPLESPERSEC;	// 1s Test
    Ton[0].accu  = 0;
    Ton[0].phase = ((U64)1750<<32) / SAMPLESPERSEC;	//234881024;
    NoOfTone = 1;
    if (codec_inactive()) {		// Codec abgeschaltet?
      codec_start();
    } else {	// Codec ist aktiv
      codec_register(CodecParam.tone_func);
      CODEC_TXINTEN();		// Run INT-Function
    } // else
  } // fi
}


void codec_toneoff(void) {
  NoOfTone = 0;
  switch(CodecState) {
    case CODECreceive:
      if (CodecParam.mutestate > 0) {	// REceive stumm?
        codec_stop();
      } else {
        codec_register(CodecParam.rx_func);
        CODEC_RXINTEN();
      }
      break;
    case CODECtransmit:			// Switch to TX
      if (codec_inactive()) {
	codec_stop();
      } else {
        codec_register(CodecParam.tx_func);
        CODEC_TXINTEN();		// Run INT-Function
      }
      break;
    default:
      codec_stop();
      break;
  } // hctiws
}




__inline tcodecstate codec_getstate(void) {
  return CodecState;
}



// calc_linearfs()
// Rechnet aus einem dB/Teilerwert -20..0dB einen Linearwert in dsp16_t 0..fast 1.0

dsp16_t calc_linearfs(U8 val, U8 divdBs) {
  float x = -(248-val) / (float)divdBs;
  return DSP16_Q(pow10f(x));
}


static const tcodecparam CodecSetupDefault = {
   // Vol	Gain	ToneVol	ToneMod	ModLimit
   DSP16_Q(0.6), DSP16_Q(0.5), DSP16_Q(1.0), DSP16_Q(0.5), DSP16_Q(0.8),
   32, 1,		// Squelch-Level, Mute-State
   &ssc_std_receive,	// Default-Receive-Int-Function
   &ssc_std_transmit,	// Default-Transmit-Int-Func
   &ssc_std_multitone 	// Default-ToneGen-Function
};


void codec_setup(tcodecsetup *Setup) {
  __int_handler new_func;
  if (Setup==NULL) {			// Standard-Werte
    CodecParam = CodecSetupDefault;
  } else {
    switch(Setup->funcsel&0x0f) {
    case 0:
      new_func = &ssc_raw_receive;
      break;
    default:
      new_func = &ssc_std_receive;
      break;
    } // hctiws RX
    if (new_func != CodecParam.rx_func) {
      CodecParam.rx_func = new_func;
      if (CodecState==CODECreceive) {
	 codec_register(new_func);
	 if (CodecParam.mutestate == 0) CODEC_RXINTEN();
      } // fi change INT-Handler
    }
    switch(Setup->funcsel>>4) {
    case 0:
      new_func = &ssc_raw_transmit;
      break;
    default:
      new_func = &ssc_std_transmit;
      break;
    } // hctiws RX
    if (new_func != CodecParam.tx_func) {
      CodecParam.tx_func = new_func;
      if (CodecState==CODECtransmit) {
	 codec_register(new_func);
	 CODEC_TXINTEN();
      } // fi change INT-Handler
    }
    CodecParam.vol  = calc_linearfs(Setup->volume, VOLUME_SCALE);
    CodecParam.gain = calc_linearfs(Setup->micgain, MICGAIN_SCALE);
    CodecParam.tone_vol = calc_linearfs(Setup->tonevol, VOLUME_SCALE);
    CodecParam.tone_mod = calc_linearfs(Setup->tonehub, MICGAIN_SCALE);
    CodecParam.modlimit = (dsp16_t)Setup->hublimit << 7; // to 16bit Signed
    CodecParam.sqlevel  = Setup->sqlevel;
  }
}


void codec_set_volume(unsigned char vol) {
  CodecParam.vol = calc_linearfs(vol, VOLUME_SCALE);
}


void codec_set_micgain(unsigned char gain) {
  CodecParam.gain = calc_linearfs(gain, MICGAIN_SCALE);
}

