/*
 * Standard DV-Modem Firmware
 * --------------------------
 * This project is firmware of the AVR32/C5 DV-Modem, developed by J.Alte.
 * It processes various converting modes for digitized speech signals.
 * See www.digisolutions.de (german only) for more infomation about DV-Modem.
 *
 * Copyright (c) 2010 by Jan Alte, DO1FJN
 *
 *
 * dv_main.c (main source file of DVfirmware)
 *
 *  Created on: 07.02.2009 (developed from DV-HW-Test)
 *      Author: Jan Alte, DO1FJN
 *
 *
 * This firmware is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This firmware is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * All source files of DVfirmware are written in UTF-8 charset.
 *
 * Report:
 * 2010-02-23 V0.80  Erstes Firmware-Release
 * 2010-03-02 V0.80a Fehlerbehebung: Initialisierungsreihenfolge SSC in Codec-Func
 * 2010-03-04 V0.81  Codec(Analog) Funktionen überarbeitet, BugFix WD,Calltone
 * 2010-03-13 V0.81a implizit Mute auch bei TX,Init, Analog-Betrieb: NFReq auch hier
 * 2010-03-14	     Nur Routen aus den SlowData werden ausgewertet (Hdr ist u.U. verä.)
 * 2010-03-16        DVOff: Ack bei TX-Aus Befehl, D-Star-Standby bei Frequenzänderung
 * 2010-03-20 V0.81b Outputgain +3dB insteed +6dB -> better match to D-Star-Volume
 *
 */

#define DVMODEMVERSION	0x0812		// Version Vxx.yy (last digit is char ' ', a, b...)

/* What can be read from this version - String?
 * V0.82c
 * Main Version  0 - This Project is in a early developing state
 * Subversion    8 - Suberversion, increased, if new features implemented
 * Subsubversion 2 - Increased if parts of the source are rewritten or optimized
 * Bugfixlevel   c - Increased with one or more bugfixes (at the same time)
 */

#include "intc.h"
#include "hw_init.h"
#include "common/bf_cmd.h"
#include "common/dstar_func.h"
#include "common/codec_func.h"

#include <string.h>

#ifndef true
#define true	1
#endif
#ifndef false
#define false	0
#endif

#define cMaxStations	10		// Speicheranzahl verschiedener empfangener Stationen
#define cPrcsAsNewTime	180		// 180s erneut gleiche Station melden

typedef enum {
  OpUnknown, Receiving, Transmitting
} tOPmode;

typedef enum {
  dvmOFF=0, dvmANALOG, dvmDSTAR, dvmHYBRID, dvmAPCO25, dvmFSK9600, dvmAFSK1200
} tDVmode;
#define MaxDefinedMode		dvmAFSK1200
#define MaxSupportedMode	dvmDSTAR	//dvmAFSK1200

typedef enum {
  dvmCHANGEREQ=0, dvmCHANGINGoff, dvmCHANGEON, dvmCHANGINGon, dvmOK
} tDVmodeSta;

typedef void (*tMDFunction)(void);	// Mode-Dependent-Function (called in main())


typedef enum {
  dmNODATA=0, dmICOMFMT, dmRAWSDATA, dmRAWBLOCK
} tDVdatamode;
#define MaxDefinedDataMode	dmRAWBLOCK
#define MaxSupportedDataMode	dmRAWBLOCK

typedef enum {
  dpSER4800=0, dpSER9600, dpSER57K6, dpSER115K2, dpUSBDEV, dpUSBOTG
} tDVdataport;
#define MaxDefinedDataPort	dpUSBOTG
#define MaxSupportedDataPort	dpSER115K2	//dpUSBOTG



#define CtrlFuncSetSize		10

typedef void (*tcontrolfunc)(void);
typedef tcontrolfunc tCtrlFuncSet[CtrlFuncSetSize];

typedef struct __attribute__((__packed__)) {
  char Route[16];
  char Call[8];
  char Sign[4];
  char ExtI[4];
  U32  Time;
} tStation;


typedef void (*tdatatxfunc)(const char *, int);
typedef int  (*tdatarxfunc)(void);
typedef int  (*tdatacopyfunc)(char *, int);


// *** KONSTANTE STRINGS ***

#ifdef HW09PT
const char DVmodemIdentStr[] = "DV-ModemHW0.9PT ";
#elif (defined HW10)
const char DVmodemIdentStr[] = "DV-ModemHW V1.0 ";
#else
const char DVmodemIdentStr[] = "DV-ModemHW V1.1 ";
#endif
// Die Kennung "DV-Modem" in obigen Strings bitte NICHT in veränderter Firmware
// verwenden. Stattdessen Rufzeichen oder Funktionskürzel verwenden.

const tString8 DVinvalidIDStr = "inval ID";

const tString8 DVModeStrings[MaxDefinedMode+1] = {
  " DV Off ", " Analog ", " D-Star ", "Auto DV ", "APCO.25 ", "FSK 9k6 ", "AFSK 1k2"
};

const tString8 DVDataModeStrings[MaxDefinedDataMode+1] = {
  "no Data ", "Icom Fmt", "RawSData", "RawBlock"
};

const tString8 DVDataPortStrings[MaxDefinedDataPort+1] = {
  "SER4800 ", "SER9600 ", "SER57K6 ", "SER115K2", "USB DEV ", "USB OTG "
};

const U32 DVBaudrates[dpSER115K2+1] = {
  4800, 9600, 57600, 115200
};


// *** Globale Variablen ***

unsigned int	RuntimeSec, OldClockCnt, NewClockCnt;

tOPmode		OpMode	  = OpUnknown;
tDVmode		DVMode	  = dvmOFF;
tDVmode		DVprevMode= dvmOFF;
tDVmodeSta	DVModeSta = dvmCHANGEREQ;
tMDFunction	ModeDependentFunction = NULL;

tbfpacket	*bfrequest;	// Zeiger auf akt. Request-Paket
tbfpacket	bfanswer;	// Buffer für Kommunikation mit C5

unsigned int	bf_rxcrcerr;

unsigned char	DVDataMode, DVDataPort;	// Daten-Config

tdatatxfunc	DVDataTransmit = NULL;	// to Serial/Usb Port (get from rx)
tdatarxfunc	DVDataReceived = NULL;
tdatacopyfunc	DVDataCopyRcvd = NULL;	// from Port (to be tx)

static char	MyCall[16];

static tStation LastRXed;	// Zuletzt empfangene Station

tStation	RXStations[cMaxStations]; // RXStations[0] = letzte/aktuellste Sta
int		RXStationFill;


// *** Station-Speicher-Funktionen ***

// Hinweis: Der Stations-Speicher ist noch nicht implementiert. DVUP 1.0a nutzt ihn nicht
// Es existiert auch noch keine Schnittstelle...

__inline int NewStationReceived(void) {
  return (memcmp(&LastRXed, &RXStations[0], 24)!=0);	// Vgl. Route+Call
}

__inline int NewCallReceived(void) {
  return (memcmp(LastRXed.Call, RXStations[0].Call, sizeof(LastRXed.Call))!=0);
}

__inline int NewRouteReceived(void) {
  return (memcmp(LastRXed.Route, RXStations[0].Route, sizeof(LastRXed.Route))!=0);
}



// *** BF BUS Kommunikation (C5) ***

void BFerror(tBFerror Error) {
  static int bf_error;
  switch (Error) {
  case BFerr_txoverflow:
  case BFerr_rxoverflow:
  case BFerr_transmit:
  case BFerr_txData:
  case BFerr_ivalidTOidle:
  case BFerr_noTxD:
  case BFerr_noACK:
    break;
  }
  bf_error++;
}


void bf_sendversion(tbfpacket *reply) {
  reply->pktlen  = 23;
  reply->data[0] = DVMODEMVERSION & 0xFF;
  reply->data[1] = (DVMODEMVERSION >> 8) & 0xFF;
  memcpy(&reply->data[2], DVmodemIdentStr, 16);
}


void no_function(void) { }

// DTMF Dual-Tone Functions

typedef U16 tdualtone[2];

const tdualtone DTMFfrequenzen[16] = {
    {1336, 941}, // 0..9
    {1209, 697}, {1336, 697}, {1477, 697},
    {1209, 770}, {1336, 770}, {1477, 770},
    {1209, 852}, {1336, 852}, {1477, 852},
    // A to D
    {1633, 697}, {1633, 770}, {1633, 852}, {1633, 941},
    // * and #
    {1209, 941}, {1477, 941}
};

const U8 DTMFcodeambe[16] = {
  0x87, 0x80, 0x84, 0x88, 0x81, 0x85, 0x89, 0x82, 0x86, 0x8A,
  0x8C, 0x8D, 0x8E, 0x8F,
  0x83, 0x8B
};


void enable_dtmfgen(U8 dtmfno, U8 tonelen) {
  if (dtmfno<16) switch(DVMode) {
  case dvmANALOG:
    codec_gentone(DTMFfrequenzen[dtmfno][0], DTMFfrequenzen[dtmfno][1], (int)tonelen*10);
    break;
  case dvmDSTAR:
    ambe_set_dtmf(DTMFcodeambe[dtmfno], (int)(tonelen+1)>>1);
    break;
  default:
    break;
  }
}



// Data Functions (RS232):

#define ICOMHEADER_MASK		0xF0
#define ICOMHEADER_TEXTID	0x30
#define ICOM_FILLBYTE		'f'

tSlowDataFrame	UserSlowData;



void DStar_SD_RawStream(const char *datatriple) {
  if (DVDataTransmit!=NULL)
    DVDataTransmit(datatriple, 3);
}


void DStar_SD_IcomMsgs(const char *datatriple) {
  static U8 _2ndlen;
  if (DVDataTransmit!=NULL) {
    if (_2ndlen==0) {		// Test Header
      if ( ((datatriple[0]&ICOMHEADER_MASK) == ICOMHEADER_TEXTID) && ((datatriple[0]&0x0F)<6) ) {
	_2ndlen = datatriple[0]&0x0F;
	if (_2ndlen < 3) {
          DVDataTransmit(datatriple+1, _2ndlen);
	  _2ndlen = 0;
	} else {
          DVDataTransmit(datatriple+1, 2);
	  _2ndlen -= 2;
	}
      } // fi Valid block type "30"
    } else {
      if ( (_2ndlen < 4) &&	// Filter Stop-Frame-Begin:
	((datatriple[0]!=0x25)||(datatriple[1]!=0x1A)||(datatriple[2]!=0xC6)) )
	 DVDataTransmit(datatriple, _2ndlen);
      _2ndlen = 0;
    }
  }
}


// Sende-Funktionen - sollten schnell abgearbeitet werden (called in int)
char *DStar_SD_IcomXmit(int cycleno) {
  if (DVDataReceived!=NULL) {
    int len = DVDataReceived();
    if (len > 5) len = 5;
    if (len > 0) {
      UserSlowData[0] = ICOMHEADER_TEXTID|len;
      memset(UserSlowData+2, ICOM_FILLBYTE, 4);	// 6 Blocks
      DVDataCopyRcvd(UserSlowData+1, len);
      return UserSlowData;
    }
  }
  return NULL;
}

char *DStar_SD_RawXmit(int cycleno) {
  if (DVDataReceived!=NULL) {
    int len = DVDataReceived();
    if (len > 0) {
      if (len > 6) len = 6;
      memset(UserSlowData+1, ICOM_FILLBYTE, 5);
      DVDataCopyRcvd(UserSlowData, len);
      return UserSlowData;
    }
  }
  return NULL;
}

char *DStar_SD_BlockXmit(int cycleno) {
  static char *BlockBufPtr;
  if (DVDataReceived!=NULL) {
    if (cycleno==1) {
      int len = DVDataReceived();
      if (len >= sizeof(UserSlowData)) {
        DVDataCopyRcvd(UserSlowData, sizeof(UserSlowData));
	BlockBufPtr = UserSlowData;
      } else BlockBufPtr = NULL;
      return BlockBufPtr;
    } else {
      char *retptr = BlockBufPtr;
      if (BlockBufPtr != NULL)
	BlockBufPtr += 6;
      return retptr;
    } // esle not start
  }
  return NULL;
}



void ChangeDataConfig(void) {
  rs232_exit();
  DVDataTransmit = NULL;
  DVDataReceived = NULL;
  DVDataCopyRcvd = NULL;
  if ((DVMode > dvmANALOG)&&((DVDataMode&0x0F) != dmNODATA)) {
    if (DVDataPort <= dpSER115K2) {	// RS232
      DVDataTransmit = rs232_transmit;
      DVDataReceived = rs232_received;
      DVDataCopyRcvd = rs232_copyblock;
      rs232_init(DVBaudrates[DVDataPort], (tFlowCtrl)((DVDataMode>>4)&0x03));
    } else {	// USB... switch(DVDataPort)...

    }
    switch (DVDataMode&0x0F) {
    case dmICOMFMT:
      dstar_setslowdatarxfct((DVDataTransmit)?DStar_SD_IcomMsgs:NULL);	// Je nach Dateneinstellung
      dstar_setslowdatatxfct((DVDataReceived)?DStar_SD_IcomXmit:NULL);
      break;
    case dmRAWSDATA:
      dstar_setslowdatarxfct((DVDataTransmit)?DStar_SD_RawStream:NULL);
      dstar_setslowdatatxfct((DVDataReceived)?DStar_SD_RawXmit:NULL);
      break;
    case dmRAWBLOCK:
      dstar_setslowdatarxfct(NULL);			// Handled in DV-Main-Function
      dstar_setslowdatatxfct((DVDataReceived)?DStar_SD_BlockXmit:NULL);
      break;
    } // hctiws
  } // fi active Data I/O
}


// Data Functions End


// Siemens C5 Steuerungsfunktionen (per BF_Bus)

void analog_receive(void) {
  codec_receive();
  bf_request_noparam(DVrpyReceive);		// Send Stop Message
}

void analog_sqclose(void) {		// Schließe Scuelch
  codec_mute();
  if (codec_rxismute())
    bf_request_noparam(DVreqNFPAoff);
}

void analog_sqopen(void) {		// Öffne Squelch
  if (codec_rxismute())			// RX-NF ist noch leise...
    bf_request_noparam(DVreqNFPAon);
  codec_unmute();
}


// *** Controlset's ***
const tCtrlFuncSet AnalogFunc = {
  //DVcmd Stop
  &codec_exit, &analog_receive, &no_function, &codec_transmit,
  // FreqChange
  &codec_mute, &analog_sqopen, &analog_sqclose,
  &codec_toneoff, &codec_calltone, &codec_toneoff
};

const tCtrlFuncSet DStarFunc = {
  &dstar_exit_hardware, &dstar_receive, &no_function, &dstar_transmit,
  // FreqChange ~ SquelchOff ~ SquelchOn
  &dstar_standby, &no_function, &no_function,
  &dstar_clr_emergency, &dstar_set_emergency, &ambe_stop_dtmf
};

const tCtrlFuncSet NothingFunc = {
  &no_function, &no_function, &no_function, &no_function,
  &no_function, &no_function, &no_function,
  &no_function, &no_function
};

const tCtrlFuncSet *BasicOpFunc[MaxDefinedMode] = {
    &AnalogFunc, &DStarFunc,
    // AutoDV, APCO.25 FSK AFSK
    &NothingFunc, &NothingFunc, &NothingFunc
};


void handle_bfpacket(void) {
  bfanswer.fromadr = BF_MYADDRESS;
  bfanswer.destadr = bfrequest->fromadr;
  bfanswer.pktlen  = 0;				// Default: Keine Antwort.
  bfanswer.cmd     = bfrequest->cmd|0x80;	// Cmd mit Bit7 gesetzt
  if (bf_getcrc(bfrequest)== 0) {		// Stimmt die CRC?
    if ((DVMode>dvmOFF) && (DVMode<=MaxSupportedMode) &&	\
      (bfrequest->cmd > 0)&&(bfrequest->cmd <= CtrlFuncSetSize)) {
      (*BasicOpFunc[DVMode-1])[(bfrequest->cmd)-1]();	// Do simple Function
    } // fi Control
    switch (bfrequest->cmd) {
      case DVcmdActivate:	// Cmd4Bootloaderto Start this app, ignore
        break;
      case DVcmdGetVers:	// sende Versionskennung, capability-Flags
        bf_sendversion(&bfanswer);
        break;
      case DVcmdReceive:
        OpMode = Receiving;
        leds_off();
        if (DVMode==dvmOFF) bf_acknowledge(&bfanswer);	// Ack TXoff
        break;
      case DVcmdTransmit:
        OpMode = Transmitting;
        bf_acknowledge(&bfanswer);
        leds_green();
	switch(DVMode) {
	case dvmDSTAR:
	  break;
	default:
	  break;
	} // hctiws
        break;
      case DVcmdFreqChge:		// Frequenz wird geändert
	// Receiver auf Standby. Wenn Senden (sollte n. vorkommen): Rx
	break;

      case DVcmdVolume:
        ambe_set_volume(bfrequest->data[0]);
        codec_set_volume(bfrequest->data[0]);
        break;
      case DVcmdMicGain:
	codec_set_micgain(bfrequest->data[0]);
	break;
      case DVcmdSquelch:
	break;
      case DVcmdDTMFTonOn:
	enable_dtmfgen(bfrequest->data[0], bfrequest->data[1]);
	break;
      case DVcmdModeChnge:
	if (bfrequest->data[0] <= MaxSupportedMode) {
	  if (DVMode != ((tDVmode)bfrequest->data[0])) {
	    DVMode = (tDVmode)bfrequest->data[0];
	    DVModeSta = dvmCHANGEREQ;
	  } // fi change
	} // fi Mode ok
	bf_sendbyte(&bfanswer, DVMode);	// Replay Current Mode
	break;
      case DVcmdMaxModeNo:
	bf_sendbyte(&bfanswer, MaxSupportedMode);
	break;
      case DVcmdModeGetStr:
	if (bfrequest->data[0] > MaxDefinedMode) {
          bf_sendstring8(&bfanswer, bfrequest->data[0], DVinvalidIDStr);
	} else {// fi Mode ok
	  bf_sendstring8(&bfanswer, bfrequest->data[0], DVModeStrings[bfrequest->data[0]]);
	}
	break;
      case DVcmdGetCurrBER:		// 0x1B
	if ((DVMode==dvmDSTAR)&&(OpMode==Receiving)) {
	  bf_sendword(&bfanswer, ambe_getber());
	} // fi relevant
	break;
      case DVcmdSetData:		// 0x1C
	// DataMode & Port Set Cmd
	DVDataMode = bfrequest->data[0];
	DVDataPort = bfrequest->data[1];
	ChangeDataConfig();
	break;
      case DVcmdGetMaxDMP:		// 0x1D
	bfanswer.pktlen = 7;
	bfanswer.data[0] = MaxSupportedDataMode;
	bfanswer.data[1] = MaxSupportedDataPort;
        break;
      case DVcmdGetDatMStr:		// 0x1E
 	if (bfrequest->data[0] > MaxDefinedDataMode) {
          bf_sendstring8(&bfanswer, bfrequest->data[0], DVinvalidIDStr);
	} else {// fi Mode ok
	  bf_sendstring8(&bfanswer, bfrequest->data[0], DVDataModeStrings[bfrequest->data[0]]);
	}
	break;
      case DVcmdGetDatPStr:		// 0x1F
 	if (bfrequest->data[0] > MaxDefinedDataPort) {
          bf_sendstring8(&bfanswer, bfrequest->data[0], DVinvalidIDStr);
	} else {// fi Mode ok
	  bf_sendstring8(&bfanswer, bfrequest->data[0], DVDataPortStrings[bfrequest->data[0]]);
	}
	break;

      case DVcmdSetMyCall:
	if (bfrequest->pktlen == (5+16)) {	// Länge ok.
	  memcpy(MyCall, bfrequest->data, 16);
	  dstar_update_mycall((char *)bfrequest->data);
	} // fi Len Ok
	break;
      case DVcmdAnalogSet:	// Komplettes Setup für Analogbetrieb
	codec_setup((tcodecsetup *)bfrequest->data);
	ambe_set_volume(bfrequest->data[1]);	// Volume auch bei D-Star
	break;
      case DVcmdSetRoute:
	dstar_update_route((char *)bfrequest->data);
	break;
      case DVcmdSetDest:
	dstar_update_dest((char *)bfrequest->data);
	break;
      case DVcmdSetDepart:
	dstar_update_depart((char *)bfrequest->data);
	break;
      case DVcmdSetCompan:
	dstar_update_yourcall((char *)bfrequest->data);
        break;
    } // hctiws
    if (bfanswer.pktlen > 4) {	// Antwort erstellt? Ja: Aussenden
      bf_sendpkt(&bfanswer);
    } // fi Antwort
  } else {		// Err
    bf_rxcrcerr++;
  }
}




// *** Modus-Abhängige Funktionen ***

#define SOFORTRXNFAUSPKTANZ	51	// 1s, Anzahl 20ms Pakete
// wenn empangender Duchgang kürzer is, schalte LS gleich aus

void DStar_CopyRouteFromHeader(tds_header *header) {
  memcpy(LastRXed.Route, header->RPT2Call, 16);	// Copy Route from Header
  memcpy(LastRXed.Call,  header->MyCall, 12);
  LastRXed.ExtI[0] = ' ';
  LastRXed.Time = RuntimeSec;
}
// abgekürzte Bedeutung der Flag 1 Bits 0..2 des D-Star Headers:
const tString8 RPTinfoStr[7] = {
  "unavail.", "no reply", " ACK    ", "resendRq", " ---    ", "a.reply ", "ctrldata"
};

void DStar_Function(void) {
  char HandleRouteReq = 0;
  tds_header RxHeader;
  char *slowdata;
  char MsgBuffer[16];
  // Rate der Aktualisierungen begrenzen!
  switch(ambe_getnewstate()) {
  case AMBEoff:
    bf_request_noparam(DVreqNFPAoff);
    break;
  case AMBEdecoding:
    bf_request_noparam(DVreqNFPAon);
    break;
  case AMBEsleep:
    if(ambe_packetcount() < SOFORTRXNFAUSPKTANZ) {
      bf_request_noparam(DVreqNFPAoff);
    } // fi
    break;
  default: // No Changing...
    break;
  } // hctiws
  if (dstar_newheader()) {		// *** Header empfangen? ***
    unsigned char RxHdr_BitErr = dstar_decodeheader(&RxHeader, (char *)dstar_getheader());
    unsigned char RptInfo = RxHeader.flags[0]&0x07;
    if (dstar_checkheader(&RxHeader)) {
      if ((RptInfo==0) &&			// Keine Flags..
	(memcmp(RxHeader.MyCall, RxHeader.RPT2Call, 6)&&	// und Call nicht Area|Dest
	 memcmp(RxHeader.MyCall, RxHeader.RPT1Call, 6)) ) {	// -> normal station
        LastRXed.ExtI[1] = (RxHdr_BitErr==0)?'*':' ';
        LastRXed.ExtI[2] = (RxHdr_BitErr<10)?'*':' ';
        LastRXed.ExtI[3] = (RxHdr_BitErr<20)?'*':' ';
        HandleRouteReq = 'H';
      } else {
	// repeater-information
	memcpy(MsgBuffer, RxHeader.MyCall, 8);
	if (RptInfo != 0) {
	  memcpy(MsgBuffer+8, RPTinfoStr[RptInfo-1], 8);
	} else {
	  memcpy(MsgBuffer+8, RxHeader.MyCall2, 4);
	  memcpy(MsgBuffer+12, " Inf", 4);
	}
	bf_send2C5(DVsendMessage, 16, MsgBuffer);
      }
    } // fi valid
  } else if (dstar_newrxSlowData()) {	// *** Neue 60Byte SlowData? ***
    slowdata = dstar_getrxSlowData();
    if (((DVDataMode&0x0F)==dmRAWBLOCK) && (DVDataTransmit!=NULL)) {
      DVDataTransmit(slowdata, sizeof(tSlowDataFrame));
    } // fi Block-Data xmit
    if (dstar_decodeheader_sd(&RxHeader, slowdata)) {
      if (dstar_checkheader(&RxHeader)) {
        LastRXed.ExtI[1] = '(';
        LastRXed.ExtI[2] = 'D';
        LastRXed.ExtI[3] = ')';
        HandleRouteReq = 'D';
      } // fi hdr ok
    } // fi else teste 4 msg/gps etc
  } // fi new slowData

  if (HandleRouteReq) {
    DStar_CopyRouteFromHeader(&RxHeader);
    if (HandleRouteReq=='D') {		// neu: Nur unveränderte Routen auswerten
      if (NewRouteReceived())
        bf_send2C5(DVsendNewRoute, sizeof(LastRXed.Route), LastRXed.Route);
    } // fi neue Route
    if ((HandleRouteReq=='H')||NewCallReceived()) {	// Header oder neue Station
      bf_send2C5(DVsendNewRxCall, 16, LastRXed.Call);
    } else if ( (LastRXed.Time-RXStations[0].Time) > cPrcsAsNewTime ) {
      // ein OM spricht immer noch / schon wieder -> Nur Anzeige auf C5-Display
      bf_send2C5(DVsendMessage, 16, LastRXed.Call);
    } // fi Inform C5

    if (NewStationReceived()) {	// Neue Empfangene Station?
      // Alte Stationen herunter schieben...
      memcpy(&RXStations[0], &LastRXed, sizeof(tStation));
      RXStationFill++;
      if (RXStationFill > cMaxStations) RXStationFill=cMaxStations;
      // Vergleich 0 mit Station 2..n? Eine gleich, dann doppelte löschen

    } // fi Einsortieren
  } // fi neue Routen-Daten
}


// *** Modus-Wechsel Funktionen ***


void Mode_ChangeOff(void) {
  switch(DVprevMode) {
  case dvmANALOG:
    codec_exit();
    break;
  case dvmDSTAR:
    dstar_exit_hardware();
    rs232_exit();
    break;
  default:
    idle_timer_stop();
    break;
  }
  ModeDependentFunction = NULL;		// deaktiviere Modus-abh. Funktionalität
}

int Mode_TestOff(void) {
  switch(DVprevMode) {
  case dvmANALOG:
    return true;
  case dvmDSTAR:
    return (ambe_getstate() <= AMBEoff);
  default:
    return true;
  }
}

void Mode_ChangeOn(void) {
  switch(DVMode) {
  case dvmANALOG:
    codec_init();
    break;
  case dvmDSTAR:
    dstar_init_hardware();		// Initialize ADC+DAC+GMSK-Timer+AMBE(SSC)
    dstar_standby();
    ModeDependentFunction = DStar_Function;
    ChangeDataConfig();
    break;
  case dvmOFF:
  default:
    if (OpMode==Transmitting) {
      bf_request_noparam(DVreqemergency);	// Sender Aus Request
      OpMode=Receiving;
    } // fi darf nicht mehr senden
    disable_afout();			// Kein Kontakt zum Modulator
    idle_timer_start();
    break;
  }
  if (DVMode!=dvmOFF) enable_afout();	// Generell durchgeschaltet wenn DV-Modem aktiviert
  if (OpMode==OpUnknown) OpMode=Receiving;
}


int Mode_TestOn(void) {
  switch(DVMode) {
  case dvmANALOG:
    if (OpMode==Transmitting)
      codec_transmit();
    else
      codec_receive();
    return true;
  case dvmDSTAR:
    if (OpMode==Transmitting) {
      dstar_transmit();
    } else
      dstar_receive();
    return true;
  default:
    return true;
  }
}



int main(void) {

#ifdef HW09PT
  U8 poti_control[2];
#endif

  Disable_global_interrupt();		// Disable all interrupts.

  // *** Initialization-Section ***
  init_hardware();
  INTC_init_interrupts();		// Initialize interrupt vectors.

  bf_init();				// BF-Bus Usart Init
  bf_set_errorfct(&BFerror);

  // *** Initialierung der verschiedenen Parameter ***
  codec_setup(NULL);			// Default-Init A
  ambe_setup();				// Prepare AMBE Buffers
  dstar_init_data();			// Default Header "noCall"

  memset(&RXStations[0], 32, sizeof(tStation));
  memset(&LastRXed, 32, sizeof(tStation));
  RXStationFill = 0;	// Keine Station empfangen


#ifdef HW09PT
  twi_init();
  poti_control[0] = 0x00;		// Volume
  poti_control[1] = 0xD0;
  twi_write(POTI_ADR, poti_control, 2);
  poti_control[0] = 0x80;		// MicGain
  poti_control[1] = 0x80;
  twi_write(POTI_ADR, poti_control, 2);
  twi_exit();
#endif

  OldClockCnt = Get_system_register(AVR32_COUNT);
  RuntimeSec  = 0;
  Enable_global_interrupt();	// Enable all interrupts.
  leds_off();			// keine LEDs an
  while (TRUE) {		// *** Main-Loop ***
#ifndef DEBUG
    SLEEP();			// erst einmal dösen bis zu einem IRQ
#endif
    watchdog();
    // Pseudo-Sekundenzähler:
    NewClockCnt = Get_system_register(AVR32_COUNT);
    if ((NewClockCnt-OldClockCnt) > MASTERCLOCK) {
      OldClockCnt = NewClockCnt;
      RuntimeSec++;
    } // fi new Second
    // Modus-Handling (Umschalten des DVMode im Betrieb)
    switch (DVModeSta) {
    case dvmCHANGEREQ:
      Mode_ChangeOff();
      DVModeSta = dvmCHANGINGoff;
      break;
    case dvmCHANGINGoff:
      if (Mode_TestOff()) DVModeSta = dvmCHANGEON;
      break;
    case dvmCHANGEON:
      Mode_ChangeOn();
      DVModeSta = dvmCHANGINGon;
      break;
    case dvmCHANGINGon:
      if (Mode_TestOn()) {
	DVModeSta = dvmOK;
	DVprevMode = DVMode;
      } // fi
      break;
    case dvmOK:
      if (ModeDependentFunction!=NULL) ModeDependentFunction();
      break;
    } // hctiws
    // BF-Bus Handling (Empfang Cmds vom C5 und Antworten senden)
    while (bf_pktreceived()) {
      bfrequest = bf_getreceivedpkt();	// Setze Zeiger auf empf. Datenpaket
      handle_bfpacket();
      bf_freereceivedpkt();
    } // ehliw
  }
}

// MAIN Ende

