/*
 * ambe_acti.c
 *
 * control and interrput functions to communicate with AMBE-2020 in active-framed mode.
 * This code ist needed by Hardware 0.9PT only and optional for newer Hardware.
 *
 *  Created on: 20.03.2009
 *      Author: Jan Alte, DO1FJN
 *
 *
 * This file is part of the DV-modem firmware (DVfirmware).
 * For general infomation about DVfirmware see "dv-main.c".
 *
 * DVfirmware is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * The DVfirmware is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Report:
 *  2009-04-04	ambe_in_sel und ambe_outsel per EPR zeigen immer auf inaktiven Buffer
 *  2009-08-02  AMBE-Start und Shutdown nun reibungslos / AMBE-Timer für Überwachung
 *
 *  ToDo:
 * -PreDefined DTMF Tone Packets
 * -easy VAD on/off define
 * -Slip-Control
 *
 */


#include "ambe_func.h"
#include "hw_defs.h"
#include "gpio_func.h"
#include "int_func.h"
#include "intc.h"
#include "compiler.h"
#include "dsp.h"	// für Aligned only
#include <string.h>	// für memcpy


#ifdef AMBEACT

#define AMBE_OPERATEID			0x0000
#define AMBE_PWRDOWNID			0x5500
#define AMBE_ID_MASK			0xFF00
#define AMBE_CTRL1_CNI_OFFSET		1
#define AMBE_CTRL1_LFI_OFFSET		7
#define AMBE_CTRL1_ID_OFFSET		8
#define AMBE_CTRL1_CNI_MASK		(1<<AMBE_CTRL1_CNI_OFFSET)
#define AMBE_CTRL1_LFI_MASK		(1<<AMBE_CTRL1_LFI_OFFSET)

#define AMBE_CTRL2_RIS_OFFSET		0
#define AMBE_CTRL2_EE_OFFSET		2
#define AMBE_CTRL2_SL_OFFSET		3
#define AMBE_CTRL2_VAD_OFFSET		5
#define AMBE_CTRL2_VOL_OFFSET		8

#define AMBE_CTRL2_RIS_BOTH		0x0
#define AMBE_CTRL2_RIS_ENCODER		0x1
#define AMBE_CTRL2_RIS_DECODER		0x2
#define AMBE_CTRL2_RIS_NONE		0x3
#define AMBE_CTRL2_RIS_MASK		0x0003

#define AMBE_CTRL2_EE_MASK		(1<<AMBE_CTRL2_EE_OFFSET)
#define AMBE_CTRL2_VAD_MASK		(1<<AMBE_CTRL2_VAD_OFFSET)

#define AMBE_TIMER_CLOCK		((MASTERCLOCK+64)/128)
#define AMBE_TIMETORESET		5	// Zeit nach Powerdown, wenn AMBE erneut starten soll
#define AMBE_TIMETORELOAD		17	// Zeit nach EPR, an dem der PDAC neu geladen wird
#define AMBE_TIMETODEATH		30	// Zeit nach EPR-obige bis AMBE für tot erklärt wird
#define AMBE_TIMETOBOOT			100	// Zeit, der der AMBE max. zum Aufwachen benötigt
#define AMBE_TIMETOLIVE			50	// Zeit vom Aufwachen bis zum ersten EPR INT


A_ALIGNED tAMBEinframe	ambe_in[2];
A_ALIGNED tAMBEoutframe	ambe_out[2];

tambestate	AMBEstate = AMBEnoinit;		// Zustand AMBE-Chip
//unsigned char	AMBEvolume;

unsigned short	BitErrorRate;

unsigned int ambe_packet_count;
unsigned int ambe_idle_count, ambe_powerdown_to, ambe_dtmf_count;

int ambe_in_sel, ambe_out_sel;

// Bit-Sync-Function (SLIP)
tbits_left_fkt	ambe_bitsleftfkt;
int		ambe_bitperiod, ambe_bitfirst, ambe_bitlast;


__inline void ambe_timer_start(void);	// Forward
void	ambe_start(void);		// Forward
void	ambe_stop(void);		// Forward
void	ambe_timer_stop(void);		// Forward


__inline void ambe_reload_pdac(void) {
  AVR32_PDCA.channel[AMBE_CHANNEL1].mar = (U32)&ambe_in[ambe_in_sel];
  AVR32_PDCA.channel[AMBE_CHANNEL1].tcr = AMBE_FRAMESIZE; // Sender zuerst, kommt gleich!
  AVR32_PDCA.channel[AMBE_CHANNEL1].cr  = AVR32_PDCA_TEN_MASK;
  AVR32_PDCA.channel[AMBE_CHANNEL0].mar = (U32)&ambe_out[ambe_out_sel^1];
  AVR32_PDCA.channel[AMBE_CHANNEL0].tcr = AMBE_FRAMESIZE;
  AVR32_PDCA.channel[AMBE_CHANNEL0].cr  = AVR32_PDCA_TEN_MASK;
}


__inline void ambe_restart_timer(void) {
  AVR32_TC.channel[AMBE_TIMER_CH].ccr = AVR32_TC_SWTRG_MASK; // | AVR32_TC_CLKEN_MASK;
}


#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ambe_timer_intPreStart(void) {	// AMBE wieder einschalten
  AVR32_TC.channel[AMBE_TIMER_CH].sr;		// Read Status to acknowledge INT
  ambe_start();
}


#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ambe_timer_intTO1(void) {	// Ende Gelände: AMBE ausschalten
  tambestate mystate = AMBEstate;	// Zustand vor "Stop" retten
  AVR32_TC.channel[AMBE_TIMER_CH].sr;	// Read Status to acknowledge INT
  ambe_stop();
  if (mystate == AMBEbooting) {
    INTC_register_interrupt(&ambe_timer_intPreStart, AMBE_TIMER_IRQ, AMBE_TIMER_PRIO);
    AVR32_TC.channel[AMBE_TIMER_CH].rc = (AMBE_TIMETORESET*AMBE_TIMER_CLOCK)/1000;
    ambe_timer_start();
  } // fi reboot
}


#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ambe_timer_intTO0(void) {
  AVR32_TC.channel[AMBE_TIMER_CH].sr;	// Read Status to acknowledge INT
  AVR32_TC.channel[AMBE_TIMER_CH].rc = (AMBE_TIMETODEATH*AMBE_TIMER_CLOCK)/1000;
  INTC_register_interrupt(&ambe_timer_intTO1, AMBE_TIMER_IRQ, AMBE_TIMER_PRIO);
  AVR32_SSC.cr = 0x0202;		// Stop both SSC Tranceiver
  AVR32_SSC.sr;				// Reset Status Register PDAC Transfer
  ambe_reload_pdac();
  AVR32_SSC.cr = 0x0101;		// Start RX and TX of SSC
}



#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ambe_epr_int(void) {
  eic_clrline(AMBE_EPR_INT);
// doppelt?  ambe_restart_timer();			// Neustart von CV=0 an
  ambe_in_sel  ^= 1;			// Toggle Paketsel
  if (AMBEstate == AMBEpowerdown) {	// Wenn Power-Down, dann nicht mehr auf weitere EPR
    eic_disableint(AMBE_EPR_INT);	// reagieren, die bei Fehlverhalten kommen könnten
  }	// fi OFF-Request
  ambe_restart_timer();			// Neustart von CV=0 an
  INTC_register_interrupt(&ambe_timer_intTO0, AMBE_TIMER_IRQ, AMBE_TIMER_PRIO);
  AVR32_TC.channel[AMBE_TIMER_CH].sr;	// Read Status to clear Flag
  AVR32_TC.channel[AMBE_TIMER_CH].rc = (AMBE_TIMETORELOAD*AMBE_TIMER_CLOCK)/1000;
  AVR32_TC.channel[AMBE_TIMER_CH].sr;	// Read Status to clear Flag
}


#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ambe_timer_intStart(void) {
  AVR32_TC.channel[AMBE_TIMER_CH].sr;	// Read Status to acknowledge INT
  AVR32_TC.channel[AMBE_TIMER_CH].rc = (AMBE_TIMETOLIVE*AMBE_TIMER_CLOCK)/1000;
  INTC_register_interrupt(&ambe_timer_intTO1, AMBE_TIMER_IRQ, AMBE_TIMER_PRIO);
  if (AMBEstate == AMBEbooting) {
    AVR32_PDCA.channel[AMBE_CHANNEL0].ier = AVR32_PDCA_IER_TRC_MASK;
    AVR32_SSC.cr = 0x0202;		// Stop both SSC Tranceiver
    AVR32_SSC.sr;			// Reset Status Register
    AVR32_SSC.cr = 0x0101;		// Start RX and TX of SSC
    ambe_packet_count = 0;
    ambe_idle_count   = 0;
    ambe_in[0].Ctrl2 |= AMBE_CTRL2_RIS_BOTH;
    ambe_in[1].Ctrl2 |= AMBE_CTRL2_RIS_BOTH;	// Übernehmen der Bitrate
    eic_clrline(AMBE_EPR_INT);
    eic_enableint(AMBE_EPR_INT, &ambe_epr_int, AMBE_EPR_PRIO, EICm_FALLING_EDGE);
    ambe_reload_pdac();			// Enable PDAC for 1st I/O
    AMBEstate = (ambe_in[ambe_in_sel].Ctrl2&AMBE_CTRL2_EE_MASK)?AMBEencoding:AMBEdecoding;
  } // fi
}



#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void ambe_transferend_int(void) {
  int ambe_last_in;
  AVR32_SSC.sr;				// Reset Status Register
  ambe_last_in = ambe_in_sel^1;
  ambe_out_sel ^= 1;
  if (ambe_out[ambe_out_sel].Header == AMBE_HEADER) {
    if ((ambe_in[ambe_last_in].Ctrl2&AMBE_CTRL2_RIS_MASK)==AMBE_CTRL2_RIS_NONE) {
      // wenn keine RATE-INFO, ist AMBE initialisiert...
      ambe_packet_count++;
      // Wenn AMBE als Decoder...
      if ((ambe_in[ambe_last_in].Ctrl2 & AMBE_CTRL2_EE_MASK) == 0) {
	switch (ambe_in[ambe_last_in].Ctrl1&(AMBE_CTRL1_LFI_MASK|AMBE_CTRL1_CNI_MASK)) {
	  case 0:	// No LFI & CNI
	    ambe_in[ambe_last_in].Ctrl1 = AMBE_OPERATEID|AMBE_CTRL1_LFI_MASK;	// LFI on
	    BitErrorRate = ambe_out[ambe_out_sel].BER;
	    break;
	  case AMBE_CTRL1_LFI_MASK:	// LFI off, CNI on
	    ambe_in[ambe_last_in].Ctrl1 = AMBE_OPERATEID|AMBE_CTRL1_CNI_MASK;	// CNI on
	    break;
	} // hctiws
      }	// fi Decoder
    } else { // fi RIS ok
      // Wenn RIS noch gesetzt... teste ob Rate mit Wunschrate übereinstimmt.
      if ((ambe_out[ambe_out_sel].Rate[0]==ambe_in[ambe_last_in].Rate[0])&&
	((ambe_out[ambe_out_sel].Ctrl2^ambe_in[ambe_last_in].Ctrl2)&0x0f)  ) { // Fehler f
         // Achtung: Umschaltung En/Decoder per EE-Bit geht nur, wenn RIS!=NONE
        ambe_in[ambe_last_in].Ctrl2 |= AMBE_CTRL2_RIS_NONE; // RIS=3: No Rate Info
        ambe_packet_count++;
      } // fi Encoderrate != gewünschte Rate
      if (ambe_dtmf_count > 0) {
        ambe_dtmf_count--;
      } else {
        LSB(ambe_in[ambe_last_in].DtmfCtrl) = 0xFF;	// sw. off DTMF-Tone
      }
    } // esle
    ambe_idle_count++;			// Idle-Counter
    if (ambe_idle_count >= ambe_powerdown_to) {
      ambe_in[ambe_in_sel].Ctrl1 = AMBE_PWRDOWNID;
      AMBEstate = AMBEpowerdown;
    } // fi PowerDownReq
  } // fi kein Datenmüll
  ambe_reload_pdac();			// Immer ein PDAC-Reload
}





void init_ambe_ssc(void) {
  AVR32_SSC.cr   = 0x00008000;		// Do a software reset
  AVR32_SSC.cmr  = 1;
  AVR32_SSC.tcmr = SSC_TCMR_INI;
  AVR32_SSC.tfmr = SSC_TFMR_INI;	// 16 Bits + POS PULSE
  AVR32_SSC.rcmr = SSC_RCMR_AMBEACT;
  AVR32_SSC.rfmr = SSC_RFMR_AMBEACT;	// RX is Input (from AMBE-STRB)
}



void ambe_init(void) {
  AMBEstate = AMBEnoinit;
  cpld_config(CPLDdisabled);
  cpld_start();
  init_ambe_ssc();
  // Der Transfer-Complete (RX) Interrupt. Hinweis: Auc TX ist dann schon fertig.
  INTC_register_interrupt(&ambe_transferend_int, AVR32_PDCA_IRQ_0+AMBE_CHANNEL0, AMBE_TRXEND_PRIO);
  AVR32_PDCA.channel[AMBE_CHANNEL0].mar = (U32)&ambe_out[0];
  AVR32_PDCA.channel[AMBE_CHANNEL0].tcr = AMBE_FRAMESIZE; // Sender zuerst, kommt gleich!
  AVR32_PDCA.channel[AMBE_CHANNEL0].tcrr= 0;
  AVR32_PDCA.channel[AMBE_CHANNEL0].mr  = 0x01 << AVR32_PDCA_SIZE_OFFSET;	// Half-Word Transfer
  AVR32_PDCA.channel[AMBE_CHANNEL0].psr = AVR32_PDCA_PID_SSC_RX;
  AVR32_PDCA.channel[AMBE_CHANNEL1].mar = (U32)&ambe_in[0];
  AVR32_PDCA.channel[AMBE_CHANNEL1].tcr = AMBE_FRAMESIZE; // Sender zuerst, kommt gleich!
  AVR32_PDCA.channel[AMBE_CHANNEL1].tcrr= 0;
  AVR32_PDCA.channel[AMBE_CHANNEL1].mr  = 0x01 << AVR32_PDCA_SIZE_OFFSET;	// Half-Word Transfer
  AVR32_PDCA.channel[AMBE_CHANNEL1].psr = AVR32_PDCA_PID_SSC_TX;
  AVR32_PDCA.channel[AMBE_CHANNEL0].cr  = AVR32_PDCA_ECLR_MASK|AVR32_PDCA_TDIS_MASK;
  AVR32_PDCA.channel[AMBE_CHANNEL1].cr  = AVR32_PDCA_ECLR_MASK|AVR32_PDCA_TDIS_MASK;
  // *** Timer ***
  AVR32_TC.channel[AMBE_TIMER_CH].cmr =  AVR32_TC_NONE << AVR32_TC_BSWTRG_OFFSET |
    AVR32_TC_NONE << AVR32_TC_BEEVT_OFFSET |
    AVR32_TC_NONE << AVR32_TC_BCPC_OFFSET |
    AVR32_TC_NONE << AVR32_TC_BCPB_OFFSET |
    AVR32_TC_NONE << AVR32_TC_ASWTRG_OFFSET |
    AVR32_TC_NONE << AVR32_TC_AEEVT_OFFSET |
    AVR32_TC_NONE << AVR32_TC_ACPC_OFFSET |
    AVR32_TC_NONE << AVR32_TC_ACPA_OFFSET |
    1 << AVR32_TC_WAVE_OFFSET |
    AVR32_TC_WAVSEL_UP_AUTO << AVR32_TC_WAVSEL_OFFSET |
    FALSE << AVR32_TC_ENETRG_OFFSET |
    AVR32_TC_EEVT_TIOB_INPUT << AVR32_TC_EEVT_OFFSET |
    AVR32_TC_EEVTEDG_NO_EDGE << AVR32_TC_EEVTEDG_OFFSET |
    FALSE << AVR32_TC_CPCDIS_OFFSET |			// Single Shoot only, Auto-Disable at RC-Match
    FALSE << AVR32_TC_CPCSTOP_OFFSET |
    AVR32_TC_BURST_NOT_GATED << AVR32_TC_BURST_OFFSET |
    0 << AVR32_TC_CLKI_OFFSET |
    AVR32_TC_TCCLKS_TIMER_CLOCK5 << AVR32_TC_TCCLKS_OFFSET;	// MASTERCLK/128
  cpld_config(CPLDcfg_ambeact);		// AMBE2020
  AMBEstate = AMBEoff;
}


void ambe_setup(void) {
  int cnt, cnt2;
  AMBEstate = AMBEnoinit;
  ambe_powerdown_to = 0xFFFFFFF0;	// Disable Power-Down TO (huge Timeout)
  ambe_bitsleftfkt  = NULL;
  ambe_packet_count = 0;
  for (cnt=0; cnt<2; cnt++) {
    ambe_in[cnt].Header = AMBE_HEADER;
    ambe_in[cnt].Rate[0] = AMBE_RATE0;
    ambe_in[cnt].Rate[1] = AMBE_RATE1;
    ambe_in[cnt].Rate[2] = AMBE_RATE2;
    ambe_in[cnt].Rate[3] = AMBE_RATE3;
    ambe_in[cnt].Rate[4] = AMBE_RATE4;
    ambe_in[cnt].Ctrl1 = (0x00<<AMBE_CTRL1_ID_OFFSET)|AMBE_CTRL1_CNI_MASK;
    ambe_in[cnt].Unused[0] = 0;
    ambe_in[cnt].Unused[1] = 0;
    ambe_in[cnt].Unused[2] = 0;
    ambe_in[cnt].DtmfCtrl = 0x00ff;	// 0dB DTMFVol, Inactive
    ambe_in[cnt].Ctrl2 = (0x80<<AMBE_CTRL2_VOL_OFFSET)|AMBE_CTRL2_EE_MASK|AMBE_CTRL2_RIS_BOTH;
    for (cnt2=0; cnt2<AMBE_FRAMESIZE/2; cnt2++)	// Clear Data
      ambe_in[cnt].Data[cnt2] = 0;
  } // fi Header-Pre-Ini
}

void ambe_powerdown(void) {
  if (AMBEstate > AMBEoff) {
    ambe_in[0].Ctrl1 = AMBE_PWRDOWNID;
    ambe_in[1].Ctrl1 = AMBE_PWRDOWNID;
    AMBEstate = AMBEpowerdown;
  } // fi
}


__inline void ambe_timer_start(void) {
  AVR32_TC.channel[AMBE_TIMER_CH].sr;
  AVR32_TC.channel[AMBE_TIMER_CH].ier = AVR32_TC_CPCS_MASK;
  AVR32_TC.channel[AMBE_TIMER_CH].ccr = AVR32_TC_SWTRG_MASK | AVR32_TC_CLKEN_MASK;
}


__inline void ambe_timer_stop(void) {  // Stop Watchdog-Timer
  AVR32_TC.channel[AMBE_TIMER_CH].ccr = AVR32_TC_CLKDIS_MASK;
  AVR32_TC.channel[AMBE_TIMER_CH].idr = 0xFF;
  AVR32_TC.channel[AMBE_TIMER_CH].sr;	// Clear Pending INT-Requests
}


void ambe_stop(void) {
  AVR32_SSC.cr  = 0x0202;		// Stop Both
  AVR32_SSC.idr = 0xFFFF;		// SSC Ints off
  cpld_config(CPLDdisabled);
  AVR32_PDCA.channel[AMBE_CHANNEL0].cr  = AVR32_PDCA_TDIS_MASK;
  AVR32_PDCA.channel[AMBE_CHANNEL1].cr  = AVR32_PDCA_TDIS_MASK;
  AVR32_PDCA.channel[AMBE_CHANNEL0].idr = AVR32_PDCA_IDR_TERR_MASK|AVR32_PDCA_IDR_TRC_MASK|AVR32_PDCA_IDR_RCZ_MASK;
  AVR32_PDCA.channel[AMBE_CHANNEL1].idr = AVR32_PDCA_IDR_TERR_MASK|AVR32_PDCA_IDR_TRC_MASK|AVR32_PDCA_IDR_RCZ_MASK;
  eic_disableint(AMBE_EPR_INT);
  ambe_timer_stop();
  AMBEstate = AMBEoff;
}


void ambe_start(void) {
  if (AMBEstate!=AMBEpowerdown) {
    cpld_config(CPLDcfg_ambeact);		// Starts AMBE2020 from Reset
    cpld_start();				// erst jetzt SSC benutzbar
    AVR32_TC.channel[AMBE_TIMER_CH].rc = (AMBE_TIMETOBOOT*AMBE_TIMER_CLOCK)/1000;
    INTC_register_interrupt(&ambe_timer_intStart, AMBE_TIMER_IRQ, AMBE_TIMER_PRIO);
    ambe_timer_start();
    ambe_in_sel  = 0;
    ambe_out_sel = 1;
  } // fi
  AMBEstate = AMBEbooting;
}



void ambe_standby(void) {
  switch(AMBEstate) {
  case AMBEdecoding:
    ambe_in[0].Ctrl1 = AMBE_OPERATEID|AMBE_CTRL1_CNI_MASK;	// CNI on
    ambe_in[1].Ctrl1 = AMBE_OPERATEID|AMBE_CTRL1_CNI_MASK;
  case AMBEencoding:
    ambe_in[0].Ctrl2 &= ~AMBE_CTRL2_EE_MASK;
    ambe_in[1].Ctrl2 &= ~AMBE_CTRL2_EE_MASK;
    break;
  default:
    return;
  } // hctiws
  AMBEstate = AMBEsleep;
}


void ambe_encode(void) {
  if ((AMBEstate != AMBEencoding)&&(AMBEstate != AMBEnoinit)) {
    ambe_in[0].Ctrl1 = AMBE_OPERATEID;
    LSB(ambe_in[0].Ctrl2) = AMBE_CTRL2_EE_MASK;
    ambe_in[1].Ctrl1 = AMBE_OPERATEID;
    LSB(ambe_in[1].Ctrl2 )= AMBE_CTRL2_EE_MASK;
    if ((AMBEstate == AMBEoff)||(AMBEstate == AMBEpowerdown))
      ambe_start();
    else if (AMBEstate == AMBEdecoding)
      AMBEstate = AMBEencoding;
  } // fi
}


void ambe_decode(void) {
  if ((AMBEstate != AMBEdecoding)&&(AMBEstate != AMBEnoinit)) {
    ambe_in[0].Ctrl1 = AMBE_OPERATEID|AMBE_CTRL1_CNI_MASK;	// CNI on
    LSB(ambe_in[0].Ctrl2) = 0;
    ambe_in[1].Ctrl1 = AMBE_OPERATEID|AMBE_CTRL1_CNI_MASK;
    LSB(ambe_in[1].Ctrl2) = 0;
    if ((AMBEstate == AMBEoff)||(AMBEstate == AMBEpowerdown))
      ambe_start();
    else if (AMBEstate == AMBEencoding)
      AMBEstate = AMBEdecoding;
  } // fi allowed states
}


void ambe_set_volume(unsigned char volume) {
  MSB(ambe_in[0].Ctrl2) = volume;
  MSB(ambe_in[1].Ctrl2) = volume;
}

void ambe_set_dtmf(unsigned char tone, int len_frames) {
  if (len_frames==0)
    ambe_dtmf_count = 0xffffffff;
  else
    ambe_dtmf_count = len_frames;
  LSB(ambe_in[0].DtmfCtrl) = tone;
  LSB(ambe_in[1].DtmfCtrl) = tone;
}

void ambe_stop_dtmf(void) {
  ambe_dtmf_count = 0;
}


void ambe_set_timeout(unsigned int time2shutdown) {
  if (time2shutdown > 0)
    ambe_powerdown_to = time2shutdown;
}


tambestate ambe_getstate(void) {
  return AMBEstate;
}


__inline int ambe_packetcount(void) {
  return ambe_packet_count;
}


unsigned short ambe_getber(void) {
  unsigned short ber = BitErrorRate;
  BitErrorRate |= 0x8000;	// LastError, must be updated with next voice pkt
  return ber;
}


void ambe_getvoice(unsigned char *dest) {
  memcpy(dest, ambe_out[ambe_out_sel].Data, AMBE_USEDDATA);
  // Untersuchen ob größerer Gap als 20ms!
  ambe_idle_count = 0;
}


void ambe_putvoice(unsigned char *voicedat) {
  int buf_sel = ambe_in_sel;
  memcpy(ambe_in[buf_sel].Data, voicedat, AMBE_USEDDATA);
  ambe_in[buf_sel].Ctrl1  =  AMBE_OPERATEID;	// CNI, LFI off
  ambe_in[buf_sel].Ctrl2 &= ~AMBE_CTRL2_EE_MASK;
  ambe_idle_count = 0;		// neue Daten - Reset Timeout
}



#endif //AMBEACT

