/*
 * bf_func.c
 *
 * API and interrupt functions of the internal serial bus communication (BF-Bus).
 * Using one USART of AVR32 µC.
 *
 *
 *  Created on: 07.03.2009
 *      Author: Jan Alte, DO1FJN
 *
 *
 *
 * This file is part of the DV-modem firmware (DVfirmware).
 * For general infomation about DVfirmware see "dv-main.c".
 *
 * DVfirmware is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * The DVfirmware is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this package. If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "bf_func.h"
#include "hw_defs.h"
#include "intc.h"

#include "compiler.h"
#include <string.h>


typedef enum {
  BFmode_idle, BFmode_busy, BFmode_receive, BFmode_transmit,
  BFmode_sendBreak, BFmode_getBus, BFmode_testBus, BFmode_haveBus
} tBFmode;

tBFerrorhandler BF_ErrorHandler;	// Fehler-Funktionshandler

tBFmode	  bf_mode;

tbfpacket bf_rxbuffer[BFRXBUFSIZE];	// Receivebuffer (eingehende Packete)
int	  bf_rxbuf_rpos, bf_rxbuf_wpos;	// Bufferselector

tbfpacket bf_receivebuf;		// Receivebuffer (Empfang)

tbfpacket bf_txbuffer[BFTXBUFSIZE];	// Transmitbuffer
int	  bf_txbuf_rpos, bf_txbuf_wpos;	// Bufferselector



void bf_reset_receive(void) {
  AVR32_PDCA.channel[BF_CHANNEL].cr  = AVR32_PDCA_ECLR_MASK|AVR32_PDCA_TDIS_MASK;
  AVR32_PDCA.channel[BF_CHANNEL].psr = BFPIDRX;
  AVR32_PDCA.channel[BF_CHANNEL].mar = (U32)&bf_receivebuf;
  AVR32_PDCA.channel[BF_CHANNEL].tcr = BF_TelegrSize;
}


void bf_set_receive(void) {
  AVR32_PDCA.channel[BF_CHANNEL].cr  = AVR32_PDCA_TEN_MASK;
  BFIO.idr  = AVR32_USART_IER_RXRDY_MASK;	// Reagiere nicht mehr auf Int-Flag
  BFIO.rtor = BFBusPktFinTO;
  bf_mode   = BFmode_receive;
}


void bf_transmit(void) {
  BFIO.cr   = AVR32_USART_CR_TXEN_MASK|AVR32_USART_CR_RXDIS_MASK|AVR32_USART_CR_STTTO_MASK;
  BFIO.idr  = AVR32_USART_IDR_RXRDY_MASK;	// Reagiere nicht mehr auf Int-Flag
  AVR32_PDCA.channel[BF_CHANNEL].cr  = AVR32_PDCA_ECLR_MASK|AVR32_PDCA_TDIS_MASK;
  AVR32_PDCA.channel[BF_CHANNEL].psr = BFPIDTX;
  AVR32_PDCA.channel[BF_CHANNEL].mar = (U32)&bf_txbuffer[bf_txbuf_rpos].fromadr;
  AVR32_PDCA.channel[BF_CHANNEL].tcr = bf_txbuffer[bf_txbuf_rpos].pktlen;
  AVR32_PDCA.channel[BF_CHANNEL].ier = AVR32_PDCA_IER_TRC_MASK|AVR32_PDCA_IER_TERR_MASK;
  AVR32_PDCA.channel[BF_CHANNEL].cr  = AVR32_PDCA_TEN_MASK;
  bf_mode = BFmode_transmit;
}


void bf_sendbreak(void) {
  BFIO.idr  = AVR32_USART_IDR_RXRDY_MASK;	// Reagiere nicht mehr auf Int-Flag
  BFIO.rtor = BFBusBreakTime;
  // Starte TimeOut und Break:
  BFIO.cr   = AVR32_USART_CR_TXEN_MASK|AVR32_USART_CR_STTBRK_MASK|
    AVR32_USART_CR_RETTO_MASK;
  bf_mode   = BFmode_sendBreak;
}


void bf_stopbreak(void) {
  // Starte TimeOut und stoppe Break:
  BFIO.rhr;				// Lese Echo
  BFIO.ier  = AVR32_USART_IER_RXRDY_MASK;
  BFIO.rtor = BFBusBreakTime;
  BFIO.cr   = AVR32_USART_CR_TXDIS_MASK|AVR32_USART_CR_STPBRK_MASK|
    AVR32_USART_CR_RETTO_MASK;
  bf_mode   = BFmode_getBus;
}


void bf_set_testBus(unsigned int TimeOut) {
  bf_reset_receive();
  BFIO.rhr;				// Lese Echo (letztes gesendetes Byte)
  BFIO.ier  = AVR32_USART_IER_RXRDY_MASK;
  BFIO.rtor = TimeOut;
  BFIO.cr   = AVR32_USART_CR_RETTO_MASK; 	// rearm TIMEOUT
  bf_mode   = BFmode_testBus;
}


void bf_set_busyBus(void) {
  bf_reset_receive();
  BFIO.rhr;				// Lese Echo (letztes gesendetes Byte)
  BFIO.ier  = AVR32_USART_IER_RXRDY_MASK;	// Bereit für RX
  BFIO.rtor = BFBusBusyTime+BFBusPktFinTO;
  BFIO.cr   = AVR32_USART_CR_RETTO_MASK; 	// rearm TIMEOUT
  bf_mode   = BFmode_busy;			// Abwarten, was kommt
}



#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void bf_handler_int(void) {
  U32 bf_usart_csr = BFIO.csr;
  // Reset Status im CSR & Start Waiting TimeOut (löscht TIMEOUT Bit in CSR)
  BFIO.cr = AVR32_USART_CR_RSTSTA_MASK|AVR32_USART_CR_STTTO_MASK;

  if (bf_usart_csr & AVR32_USART_CSR_RXBRK_MASK) {	// End Of Break = BusReq
    switch(bf_mode) {
    case BFmode_sendBreak:
      bf_stopbreak();					// Wechsel nach "getBus"
      break;
    case BFmode_getBus:
      bf_set_testBus(BFBusPktFinTO+BFBusTestTime+BFBusPktEarlier);
      break;
    case BFmode_transmit:
    default:	// Break von einem anderen Einheit (Hörer, C5-Master)
      bf_set_busyBus();
      break;
    } // hctiws
  } else if (bf_usart_csr & AVR32_USART_CSR_RXRDY_MASK) { // Zeichen empfangen?
    switch (bf_mode) {
    case BFmode_transmit:				// *** Ende Transmission + TO ***
      bf_txbuf_rpos = (bf_txbuf_rpos+1) % BFTXBUFSIZE;
      bf_set_testBus(BFBusPktFinTO+BFBusPktEarlier+BFBusTestTime);
    default:
      bf_reset_receive();
      bf_set_receive();	// bf_mode wechselt auf RX!
      break;
    }
  } else if (bf_usart_csr & AVR32_USART_CSR_TIMEOUT_MASK) {
    switch (bf_mode) {
    case BFmode_busy:					// *** No Data after Alien-Break **
      if (bf_txbuf_rpos != bf_txbuf_wpos) {		// Ist was da zum Senden?
	bf_sendbreak();					// Sende eigenen Break...
      } else {						// sonst
        bf_mode = BFmode_idle;				// idle
      }
      break;
    case BFmode_receive:				// *** Received Data ***
      bf_set_testBus(BFBusTestTime+BFBusPktEarlier);
      if ((bf_receivebuf.destadr==BF_MYADDRESS)) {	// Filter Packets
	if (bf_receivebuf.pktlen <= BF_TelegrSize) {	// Copy valid Packet
	  memcpy(&bf_rxbuffer[bf_rxbuf_wpos], &bf_receivebuf, (size_t)bf_receivebuf.pktlen);
	  bf_rxbuf_wpos = (bf_rxbuf_wpos+1) % BFRXBUFSIZE;
	  if (bf_rxbuf_wpos==bf_rxbuf_rpos) {
	    bf_rxbuf_rpos = (bf_rxbuf_wpos+1) % BFRXBUFSIZE;
	    if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_rxoverflow);
	  } // fi rxBufferOverflow
	} // fi Len ok
      } // fi a packet for me
      break;
    case BFmode_testBus:				// BFMode 2 -> Senden innerhab der TO-Zeit
      if (bf_txbuf_rpos != bf_txbuf_wpos) {		// Ist was da zum Senden?
	bf_transmit();					// Beginne Aussendung.
      } else {						// sonst
	BFIO.rtor = BFBusBusyTime;			// ...15.96ms lang
	BFIO.cr   = AVR32_USART_CR_RETTO_MASK;		// rearm TIMEOUT
        bf_mode   = BFmode_haveBus;			// Senden ohne Break möglich...
      } // esle
      break;
    case BFmode_haveBus:				// BFMode 3 Ende -> Senden nur noch nach BusAck
      BFIO.rtor = BFBusPktFinTO;
      bf_mode   = BFmode_idle;
      break;
    // Üblicherweise nicht erreichbare Zustände:
    case BFmode_idle:					// *** Idle - Bus erst nach Break 'scharf' ***
      if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_ivalidTOidle);
      break;
    case BFmode_transmit:				// *** Ende Transmission + TO ***
      // Eigenes Last-Byte nicht empfangen
      // Mögliche Fehler: TX nicht angeschlossen, RX strong High
      bf_set_testBus(BFBusPktFinTO+BFBusPktEarlier+BFBusTestTime);
      if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_noTxD);
      break;
    case BFmode_sendBreak:	// Break Timeout abgelaufen
       // Mögliche Fehler: TX nicht angeschlossen, RX strong High
       bf_stopbreak();		// beende Break
      if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_noTxD);
      break;
    case BFmode_getBus:					// Master (C5) gibt kein ACK!
      if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_noACK);
 //     bf_sendbreak();		// try again
      break;
    } // hctiws
  } // fi TIMEOUT
}



#if __GNUC__
__attribute__((__interrupt__))
#elif __ICCAVR32__
__interrupt
#endif
static void bf_transferrdy_int(void) {
  // gleich Int Abschalten (Single Shot Like):
  AVR32_PDCA.channel[BF_CHANNEL].idr = AVR32_PDCA_IDR_TERR_MASK|AVR32_PDCA_IDR_TRC_MASK|AVR32_PDCA_IDR_RCZ_MASK;
  if (AVR32_PDCA.channel[BF_CHANNEL].isr&AVR32_PDCA_ISR_TERR_MASK) {
    if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_transmit);
  } //esle alles gut
  BFIO.cr   = AVR32_USART_CR_RXEN_MASK|AVR32_USART_CR_TXDIS_MASK|AVR32_USART_CR_RSTSTA_MASK|AVR32_USART_CR_STTTO_MASK;
  BFIO.rtor = 1; //BFByteTime;	// Echo-Byte abwarten, Timeout schaltet in den Testmode
}




unsigned char bf_pktreceived(void) {
  return bf_rxbuf_wpos != bf_rxbuf_rpos;
}


__inline tbfpacket *bf_getreceivedpkt(void) {
  return &bf_rxbuffer[bf_rxbuf_rpos];
}


void bf_freereceivedpkt(void) {
  if (bf_rxbuf_wpos != bf_rxbuf_rpos) {
    bf_rxbuf_rpos = (bf_rxbuf_rpos+1) % BFRXBUFSIZE;
  }
}


void bf_copyreceivedpkt(tbfpacket *dest) {
  if (bf_rxbuf_wpos != bf_rxbuf_rpos) {
    memcpy(dest, &bf_rxbuffer[bf_rxbuf_rpos], sizeof(tbfpacket));
    bf_rxbuf_rpos = (bf_rxbuf_rpos+1) % BFRXBUFSIZE;
  } // fi
}


void bf_init(void) {
  BFIO.cr = AVR32_USART_CR_RXDIS_MASK|AVR32_USART_CR_TXDIS_MASK|
    AVR32_USART_CR_STPBRK_MASK|AVR32_USART_CR_RSTRX_MASK|AVR32_USART_CR_RSTTX_MASK;
  AVR32_PDCA.channel[BF_CHANNEL].mr  = 0x00 << AVR32_PDCA_SIZE_OFFSET;	// Byte Transfer
  AVR32_PDCA.channel[BF_CHANNEL].tcrr = 0;				// No Next Buffer
  INTC_register_interrupt(&bf_transferrdy_int, AVR32_PDCA_IRQ_0+BF_CHANNEL, AVR32_INTC_INT0);
  bf_reset_receive();
  BF_ErrorHandler = NULL;
  bf_mode = BFmode_idle;
  bf_rxbuf_wpos = 0;
  bf_rxbuf_rpos = 0;
  bf_txbuf_wpos = 0;
  bf_txbuf_rpos = 0;
  BFIO.mr = (AVR32_USART_MR_NBSTOP_2 << AVR32_USART_MR_NBSTOP_OFFSET) | // 2 StopBits
    (AVR32_USART_MR_PAR_NONE << AVR32_USART_MR_PAR_OFFSET) |	// no Parity
    ((8 - 5) << AVR32_USART_MR_CHRL_OFFSET) |	// char length = 8
    (AVR32_USART_MR_MODE_NORMAL << AVR32_USART_MR_MODE_OFFSET);	// automatic Echo
  BFIO.brgr = (MASTERCLOCK+(8*BF_BAUDRATE))/(16*BF_BAUDRATE);
  BFIO.rtor = BFBusPktFinTO;	// Receive TimeOut Register
  // Init BF-Handler:
  INTC_register_interrupt(&bf_handler_int, BFIRQ, AVR32_INTC_INT0);
  BFIO.ier = AVR32_USART_IER_RXBRK_MASK|AVR32_USART_IER_TIMEOUT_MASK|AVR32_USART_IER_RXRDY_MASK;
  BFIO.cr  = AVR32_USART_CR_RXEN_MASK;
}


void bf_set_errorfct(tBFerrorhandler ErrorFunction) {
  BF_ErrorHandler = ErrorFunction;
}


unsigned char bf_getcrc(const tbfpacket *Packet) {
  unsigned char crc;
  int cnt;
  if (Packet->pktlen > BF_TelegrSize) return 0xff;
  crc = Packet->fromadr ^ Packet->destadr ^ Packet->pktlen ^ Packet->cmd;
  for (cnt=0; cnt < Packet->pktlen-4; cnt++) {
    crc ^= Packet->data[cnt];
  } // rof
  return crc;
}



void bf_sendpkt(const tbfpacket *Packet) {
  unsigned char crc = 0;
  unsigned char *pktdata;
  int cnt, wpos;
  if (Packet->pktlen > BF_TelegrSize) {
    if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_txData);
    return;
  } // fi Error Data
  wpos = bf_txbuf_wpos;				// make a copy 4 integraty
  pktdata = (unsigned char *)&bf_txbuffer[wpos];
  memcpy(pktdata, Packet, (size_t)(Packet->pktlen)-1);
  for (cnt=0; cnt < ((Packet->pktlen)-1); cnt++) {
    crc ^= *(pktdata++);
  } // rof
  *pktdata = crc;
  cnt = (wpos+1) % BFTXBUFSIZE;			// Next Buffer -> cnt (temp var)
  if (cnt != bf_txbuf_rpos) {			// no overrun
    bf_txbuf_wpos = cnt;
    // Vergleiche alte Schreib-Pos mit Lese-Pos, wenn gleich: Sender idle
    if (wpos == bf_txbuf_rpos) {		// idle, nothing 2 tx
      switch (bf_mode) {
      case BFmode_idle:
	bf_sendbreak();
	break;
      case BFmode_haveBus:
	bf_transmit();
	break;
      default:	// Tue nix.
	break;
      } // hctiws
    } // fi idle
  } else {	// Overrun!
    if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_txoverflow);
  } // fi no ovr
}



void bf_send2C5(char Cmd, char Len, const char *Data) {
  unsigned char crc = 0;
  char *wpkt;
  int cnt, wpos;
  if (Len > (BF_TelegrSize-5)) {
    if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_txData);
    return;
  } // fi Error Data
  wpos = bf_txbuf_wpos;				// make a copy 4 integraty
  wpkt = (char *)&bf_txbuffer[wpos];
  wpkt[0] = BF_MYADDRESS;
  wpkt[1] = BF_MASTER;
  wpkt[2] = Len+5;
  wpkt[3] = Cmd;
  memcpy(wpkt+4, Data, Len);
  for (cnt=0; cnt < (Len+4); cnt++) {
    crc ^= *(wpkt++);
  } // rof
  *wpkt = crc;
  cnt = (wpos+1) % BFTXBUFSIZE;			// Next Buffer -> cnt (temp var)
  if (cnt != bf_txbuf_rpos) {			// no overrun
    bf_txbuf_wpos = cnt;
    // Vergleiche alte Schreib-Pos mit Lese-Pos, wenn gleich: Sender idle
    if (wpos == bf_txbuf_rpos) {		// idle, nothing 2 tx
      switch (bf_mode) {
      case BFmode_idle:
	bf_sendbreak();
	break;
      case BFmode_haveBus:
	bf_transmit();
	break;
      default:	// Tue nix.
	break;
      } // hctiws
    } // fi idle
  } else {	// Overrun!
    if (BF_ErrorHandler!=NULL) BF_ErrorHandler(BFerr_txoverflow);
  } // fi no ovr
}

